from onegov.activity.collections.activity import ActivityCollection
from onegov.activity.collections.booking import BookingCollection
from onegov.activity.collections.occasion import OccasionCollection
from onegov.activity.collections.period import PeriodCollection

__all__ = [
    'ActivityCollection',
    'BookingCollection',
    'OccasionCollection',
    'PeriodCollection'
]



from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from sqlalchemy import Boolean
from sqlalchemy import CheckConstraint
from sqlalchemy import column
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import Index
from sqlalchemy import Text
from sqlalchemy.orm import object_session, relationship
from uuid import uuid4


class Period(Base, TimestampMixin):

    __tablename__ = 'periods'

    #: The public id of this period
    id = Column(UUID, primary_key=True, default=uuid4)

    #: The public title of this period
    title = Column(Text, nullable=False)

    #: Only one period is active at a time
    active = Column(Boolean, nullable=False, default=False)

    #: Start of the wishlist-phase
    prebooking_start = Column(Date, nullable=False)

    #: End of the wishlist-phase
    prebooking_end = Column(Date, nullable=False)

    #: Date of the earliest possible occasion start of this period
    execution_start = Column(Date, nullable=False)

    #: Date of the latest possible occasion end of this period
    execution_end = Column(Date, nullable=False)

    __table_args__ = (
        CheckConstraint((
            '"prebooking_start" <= "prebooking_end" AND '
            '"prebooking_end" <= "execution_start" AND '
            '"execution_start" <= "execution_end"'
        ), name='period_date_order'),
        Index(
            'only_one_active_period', 'active',
            unique=True, postgresql_where=column('active') == True
        )
    )

    #: The occasions linked to this period
    occasions = relationship(
        'Occasion',
        order_by='Occasion.start',
        backref='period'
    )

    #: The bookings linked to this period
    bookings = relationship(
        'Booking',
        backref='period'
    )

    def activate(self):
        """ Activates the current period, causing all occasions and activites
        to update their status and book-keeping.

        It also makes sure no other period is active.

        """
        if self.active:
            return

        session = object_session(self)
        model = self.__class__

        active_period = session.query(model)\
            .filter(model.active == True).first()

        if active_period:
            active_period.deactivate()

        # avoid triggering the only_one_active_period index constraint
        session.flush()

        self.active = True

    def deactivate(self):
        """ Deactivates the current period, causing all occasions and activites
        to update their status and book-keeping.

        """

        if not self.active:
            return

        self.active = False



from onegov.activity.models import Activity, Booking, Occasion, Period
from onegov.activity.collections import (
    ActivityCollection,
    BookingCollection,
    OccasionCollection,
    PeriodCollection
)


__all__ = [
    'Activity',
    'Booking',
    'Occasion',
    'Period',
    'ActivityCollection',
    'BookingCollection',
    'OccasionCollection',
    'PeriodCollection'
]



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""

from onegov.activity.models import Activity, Occasion
from onegov.core.orm.types import UUID
from onegov.core.upgrade import upgrade_task
from psycopg2.extras import NumericRange
from sqlalchemy import Column, Text, Integer
from sqlalchemy.dialects.postgresql import ARRAY, INT4RANGE


@upgrade_task('Redesign occasion table')
def redesign_occasion_table(context):
    context.operations.drop_column('occasions', 'booking_start')
    context.operations.drop_column('occasions', 'min_age')
    context.operations.drop_column('occasions', 'max_age')
    context.operations.drop_column('occasions', 'spots')

    context.operations.add_column('occasions', Column(
        'note', Text, nullable=True))

    context.operations.add_column('occasions', Column(
        'age', INT4RANGE, nullable=True))

    context.operations.add_column('occasions', Column(
        'spots', INT4RANGE, nullable=True))

    for occasion in context.session.query(Occasion).all():
        occasion.age = NumericRange(6, 17, bounds='[]')
        occasion.spots = NumericRange(0, 10, bounds='[]')

    context.session.flush()

    context.operations.alter_column('occasions', 'spots', nullable=False)
    context.operations.alter_column('occasions', 'age', nullable=False)


@upgrade_task('Make occasion location optional')
def make_occasion_location_optional(context):
    context.operations.alter_column('occasions', 'location', nullable=True)


@upgrade_task('Ensure occasions/bookings cannot be orphaned')
def ensure_occasions_bookings_cannot_be_orphaned(context):
    context.operations.alter_column('occasions', 'activity_id', nullable=False)
    context.operations.alter_column('bookings', 'occasion_id', nullable=False)


@upgrade_task('Add reporter column')
def add_reporter_column(context):
    context.operations.add_column('activities', Column(
        'reporter', Text, nullable=True))

    for activity in context.session.query(Activity).all():
        activity.reporter = activity.username

    context.session.flush()

    context.operations.alter_column('activities', 'reporter', nullable=True)


@upgrade_task('Add archived state to occasions')
def add_archive_state(context):

    # ALTER TYPE statements don't work inside of transactions -> if you reuse
    # this code, do not put it inside an update that has other things going
    # on!
    connection = context.operations.get_bind()
    previous_isolation_level = connection.get_isolation_level()
    connection.execution_options(isolation_level='AUTOCOMMIT')

    try:
        context.operations.execute(
            "ALTER TYPE occasion_state ADD VALUE IF NOT EXISTS 'archived'")
    finally:
        connection.execution_options(isolation_level=previous_isolation_level)


@upgrade_task('Add start before end constraint to occasions')
def add_start_before_end_constraint_to_occasions(context):

    context.operations.create_check_constraint(
        "start_before_end",
        "occasions",
        '"start" <= "end"'
    )


@upgrade_task('Add occasion durations')
def add_occasion_durations(context):

    context.operations.add_column(
        'activities', Column('durations', Integer, nullable=True, default=0))

    # force an update of all occasions
    for occasion in context.session.query(Occasion).all():
        occasion.note = occasion.note and occasion.note + ' ' or ' '

    context.session.flush()


@upgrade_task('Add occasion durations (second step)')
def add_occasion_durations_second_step(context):

    # undo the damage from the last step
    for occasion in context.session.query(Occasion).all():
        occasion.note = occasion.note and occasion.note.strip() or None

    context.session.flush()


@upgrade_task('Add activity ages')
def add_activity_ages(context):

    context.operations.add_column(
        'activities', Column('ages', ARRAY(INT4RANGE), default=list))

    # force an update of all occasions
    for occasion in context.session.query(Occasion).all():
        occasion.note = occasion.note and occasion.note + ' ' or ' '

    context.session.flush()


@upgrade_task('Add activity ages (second step)')
def add_activity_ages_second_step(context):

    # undo the damage from the last step
    for occasion in context.session.query(Occasion).all():
        occasion.note = occasion.note and occasion.note.strip() or None

    context.session.flush()


@upgrade_task('Rebuild models')
def rebuild_models(context):
    # delete the existing occasions/bookings table and start anew ->
    # too many changes + we're at an early stage
    context.operations.drop_table('bookings')
    context.operations.drop_table('occasions')
    context.operations.add_column(
        'activities', Column('period_ids', ARRAY(UUID), default=list))


@upgrade_task('Drop occasion state type')
def drop_occasion_state_type(context):
    context.operations.execute('DROP TYPE occasion_state')



from onegov.activity.models.occasion import Occasion
from onegov.core.orm import Base
from onegov.core.orm.mixins import (
    content_property,
    ContentMixin,
    meta_property,
    TimestampMixin,
)
from onegov.core.orm.types import UUID
from onegov.core.utils import normalize_for_url
from onegov.user import User
from sqlalchemy import Column, Enum, Text, ForeignKey, Integer
from sqlalchemy import func, distinct
from sqlalchemy.dialects.postgresql import HSTORE, ARRAY
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import relationship
from sqlalchemy_utils import aggregated, observes, IntRangeType
from uuid import uuid4


# Note, a database migration is needed if these states are changed
ACTIVITY_STATES = ('preview', 'proposed', 'accepted', 'denied', 'archived')


class Activity(Base, ContentMixin, TimestampMixin):
    """ Describes an activity that is made available to participants on
    certain occasions (i.e. dates).

    The activity describes the what's going on, the occasion describes when
    and with whom.

    """

    __tablename__ = 'activities'

    #: An internal id for references (not public)
    id = Column(UUID, primary_key=True, default=uuid4)

    #: A nice id for the url, readable by humans
    name = Column(Text, nullable=False, unique=True)

    #: The title of the activity
    title = Column(Text, nullable=False)

    #: The normalized title for sorting
    order = Column(Text, nullable=False, index=True)

    #: Describes the activity briefly
    lead = meta_property('lead')

    #: Describes the activity in detail
    text = content_property('text')

    #: Tags/Categories of the activity
    _tags = Column(MutableDict.as_mutable(HSTORE), nullable=True, name='tags')

    #: The user to which this activity belongs to (organiser)
    username = Column(Text, ForeignKey(User.username), nullable=False)

    #: The user which initially reported this activity (same as username, but
    #: this value may not change after initialisation)
    reporter = Column(Text, nullable=False)

    #: Access the user linked to this activity
    user = relationship('User')

    @aggregated('occasions', Column(Integer, default=0))
    def durations(self):
        return func.sum(distinct(Occasion.duration))

    @aggregated('occasions', Column(ARRAY(IntRangeType), default=list))
    def ages(self):
        return func.array_agg(distinct(Occasion.age))

    @aggregated('occasions', Column(ARRAY(UUID), default=list))
    def period_ids(self):
        return func.array_agg(distinct(Occasion.period_id))

    #: The occasions linked to this activity
    occasions = relationship(
        'Occasion',
        order_by='Occasion.start',
        backref='activity'
    )

    #: the type of the item, this can be used to create custom polymorphic
    #: subclasses of this class. See
    #: `<http://docs.sqlalchemy.org/en/improve_toc/\
    #: orm/extensions/declarative/inheritance.html>`_.
    type = Column(Text, nullable=True)

    #: the state of the activity
    state = Column(
        Enum(*ACTIVITY_STATES, name='activity_state'),
        nullable=False,
        default='preview'
    )

    __mapper_args__ = {
        'polymorphic_on': 'type',
        'order_by': order,
    }

    @observes('title')
    def title_observer(self, title):
        self.order = normalize_for_url(title)

    @observes('username')
    def username_observer(self, username):
        if not self.reporter:
            self.reporter = username

    @property
    def tags(self):
        return set(self._tags.keys()) if self._tags else set()

    @tags.setter
    def tags(self, value):
        self._tags = {k: '' for k in value} if value else None

    def propose(self):
        assert self.state == 'preview'
        self.state = 'proposed'

        return self

    def accept(self):
        assert self.state == 'proposed'
        self.state = 'accepted'

        return self

    def deny(self):
        assert self.state == 'proposed'
        self.state = 'denied'

        return self

    def archive(self):
        assert self.state == 'accepted'
        self.state = 'archived'

        return self



from onegov.activity.utils import merge_ranges


def test_merge_ranges():
    assert merge_ranges([[1, 2]]) == [(1, 2)]
    assert merge_ranges([[1, 2], [2, 3]]) == [(1, 3)]
    assert merge_ranges([[1, 2], [2, 3]]) == [(1, 3)]
    assert merge_ranges([[1, 2], [3, 4]]) == [(1, 2), (3, 4)]
    assert merge_ranges([[1, 2], [1, 2], [2, 3]]) == [(1, 3)]



from onegov.activity.models import Period
from onegov.core.collection import GenericCollection


class PeriodCollection(GenericCollection):

    @property
    def model_class(self):
        return Period

    def add(self, title, prebooking, execution, active=False):

        return super().add(
            title=title,
            prebooking_start=prebooking[0],
            prebooking_end=prebooking[1],
            execution_start=execution[0],
            execution_end=execution[1],
            active=active
        )

    def active(self):
        return self.query().filter(Period.active == True).first()



import string

from random import SystemRandom


GROUP_CODE_CHARS = string.ascii_uppercase + string.digits
GROUP_CODE_LENGTH = 10


def random_group_code():
    random = SystemRandom()

    return ''.join(
        random.choice(GROUP_CODE_CHARS)
        for _ in range(GROUP_CODE_LENGTH)
    )


def overlaps(range_a, range_b):
    return range_b[0] <= range_a[0] and range_a[0] <= range_b[1] or\
        range_a[0] <= range_b[0] and range_b[0] <= range_a[1]


def merge_ranges(ranges):
    """ Merges the given list of ranges into a list of ranges including only
    exclusive ranges. The ranges are turned into tuples to make them
    hashable.

    """

    ranges = sorted(list(ranges))

    # stack of merged values
    merged = [tuple(ranges[0])]

    for r in ranges:
        if overlaps(merged[-1], r):
            merged[-1] = (merged[-1][0], r[1])
        else:
            merged.append(tuple(r))

    return merged



import pytest
import sqlalchemy
import transaction

from datetime import datetime
from onegov.activity import ActivityCollection
from onegov.activity import Occasion
from onegov.activity import OccasionCollection
from onegov.activity import PeriodCollection
from onegov.activity.models import DAYS
from onegov.activity.models import Booking
from pytz import utc
from sedate import replace_timezone


def test_add_activity(session, owner):

    collection = ActivityCollection(session)
    collection.add(
        title="Visit the Butcher",
        username=owner.username,
        lead="Come visit the butcher with us and kill your own baby pig",
        text="<h1>Babe was such an overrated movie</h1>",
        tags=('butcher', 'killing', 'blood', 'fun'),
    )

    activity = collection.by_name('visit-the-butcher')
    assert activity.title == "Visit the Butcher"
    assert activity.username == owner.username
    assert activity.reporter == owner.username
    assert activity.lead.startswith("Come visit the butcher")
    assert activity.text.startswith("<h1>Babe was such a")
    assert activity.tags == {'butcher', 'killing', 'blood', 'fun'}


def test_unique_activity(session, owner):

    collection = ActivityCollection(session)
    assert collection.get_unique_name("Möped Lads") == 'moeped-lads'

    collection.add("Möped Lads", username=owner.username)
    assert collection.get_unique_name("Möped Lads") == 'moeped-lads-1'

    collection.add("Möped Lads", username=owner.username)
    assert collection.get_unique_name("Möped Lads") == 'moeped-lads-2'


def test_activity_pagination(session, owner):

    collection = ActivityCollection(session)

    for i in range(0, 20):
        collection.add(
            title='{:02d}'.format(i),
            username=owner.username
        )

    collection = collection.page_by_index(0)
    assert [a.title for a in collection.batch] == [
        "00", "01", "02", "03", "04", "05", "06", "07", "08", "09"
    ]

    collection = collection.page_by_index(1)
    assert [a.title for a in collection.batch] == [
        "10", "11", "12", "13", "14", "15", "16", "17", "18", "19"
    ]


def test_activity_order(session, owner):

    collection = ActivityCollection(session)
    collection.add(title="Ä", username=owner.username)
    collection.add(title="B", username=owner.username)
    collection.add(title="C", username=owner.username)

    assert [a.title for a in collection.query().all()] == ["Ä", "B", "C"]


def test_activity_states(session, owner):

    c = ActivityCollection(session)
    c.add("A", username=owner.username)
    c.add("B", username=owner.username).propose()
    c.add("C", username=owner.username).propose().accept()
    c.add("D", username=owner.username).propose().deny()
    c.add("E", username=owner.username).propose().accept().archive()

    c.states = ('preview', )
    assert c.query().count() == 1

    c.states = ('preview', 'proposed')
    assert c.query().count() == 2

    c.states = ('preview', 'proposed', 'accepted')
    assert c.query().count() == 3

    c.states = ('preview', 'proposed', 'accepted', 'denied')
    assert c.query().count() == 4

    c.states = ('preview', 'proposed', 'accepted', 'denied', 'archived')
    assert c.query().count() == 5

    c.states = None
    assert c.query().count() == 5


def test_activity_used_tags(session, owner):

    c = ActivityCollection(session)

    activities = (
        c.add("Paintball", username=owner.username, tags=('sport', 'fun')),
        c.add("Dancing", username=owner.username, tags=('dance', 'fun'))
    )

    assert c.used_tags == {'sport', 'dance', 'fun'}

    c.states = ('proposed', )

    assert c.used_tags == {'sport', 'dance', 'fun'}

    activities[0].propose()

    assert c.used_tags == {'sport', 'dance', 'fun'}

    class LimitedActivityCollection(ActivityCollection):

        def query_base(self):
            return self.session.query(self.model_class).filter(
                self.model_class.state == 'proposed')

    transaction.commit()

    c = LimitedActivityCollection(session)
    assert c.used_tags == {'sport', 'fun'}


def test_occasions(session, owner):

    activities = ActivityCollection(session)
    occasions = OccasionCollection(session)
    periods = PeriodCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True
    )

    tournament = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        location="Lucerne",
        age=(6, 9),
        spots=(2, 10),
        note="Bring game-face",
        activity=activities.add("Sport", username=owner.username),
        period=period
    )

    transaction.commit()

    tournament = occasions.query().one()
    assert tournament.start == datetime(2016, 10, 4, 11, tzinfo=utc)
    assert tournament.end == datetime(2016, 10, 4, 12, tzinfo=utc)
    assert tournament.timezone == "Europe/Zurich"
    assert tournament.location == "Lucerne"
    assert tournament.note == "Bring game-face"
    assert tournament.activity_id == activities.query().first().id
    assert tournament.period_id == periods.query().first().id
    assert tournament.active

    # postgres ranges are coeced into half-open intervals
    assert tournament.age.lower == 6
    assert tournament.age.upper == 10
    assert tournament.spots.lower == 2
    assert tournament.spots.upper == 11

    transaction.commit()

    tournament = occasions.query().first()
    assert tournament.age.lower == 6
    assert tournament.age.upper == 10
    assert tournament.spots.lower == 2
    assert tournament.spots.upper == 11

    # the occasions active flag is bound to the period's activity
    periods.query().first().deactivate()
    transaction.commit()
    assert not occasions.query().first().active

    periods.query().first().activate()
    transaction.commit()
    assert occasions.query().first().active


def test_profiles(session, owner):

    activities = ActivityCollection(session)
    occasions = OccasionCollection(session)
    periods = PeriodCollection(session)

    sport = activities.add("Sport", username=owner.username)

    autumn = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31))
    )

    winter = periods.add(
        title="Winter 2016",
        prebooking=(datetime(2016, 11, 1), datetime(2016, 11, 30)),
        execution=(datetime(2016, 12, 1), datetime(2016, 12, 31))
    )

    autumn_id, winter_id = autumn.id, winter.id

    # 2 occasions in autumn
    occasions.add(
        start=datetime(2016, 10, 1, 13),
        end=datetime(2016, 10, 1, 14),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=sport,
        period=autumn
    )

    occasions.add(
        start=datetime(2016, 10, 1, 13),
        end=datetime(2016, 10, 1, 14),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=sport,
        period=autumn
    )

    # 1 occasion in winter
    occasions.add(
        start=datetime(2016, 12, 1, 13),
        end=datetime(2016, 12, 1, 14),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=sport,
        period=winter
    )

    transaction.commit()

    sport = activities.query().first()
    assert len(sport.period_ids) == 2
    assert winter_id in sport.period_ids
    assert autumn_id in sport.period_ids

    # drop the winter occasion
    occasions.delete(
        occasions.query().filter(Occasion.period_id == winter_id).first())
    transaction.commit()

    sport = activities.query().first()
    assert len(sport.period_ids) == 1
    assert autumn_id in sport.period_ids


def test_occasions_daterange_constraint(session, owner):
    sport = ActivityCollection(session).add("Sport", username=owner.username)
    sport.occasions.append(Occasion(
        start=replace_timezone(datetime(2020, 10, 10), 'Europe/Zurich'),
        end=replace_timezone(datetime(2010, 10, 10), 'Europe/Zurich'),
        timezone='Europe/Zurich'
    ))

    with pytest.raises(sqlalchemy.exc.IntegrityError):
        session.flush()


def test_no_orphan_bookings(session, owner):

    activities = ActivityCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)

    tournament = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=activities.add("Sport", username=owner.username),
        period=periods.add(
            title="Autumn 2016",
            prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
            execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
            active=True
        )
    )

    tournament.bookings.append(Booking(
        username=owner.username,
        last_name='Muster',
        first_name='Peter',
        period_id=periods.query().first().id
    ))

    session.flush()

    with pytest.raises(sqlalchemy.exc.IntegrityError):
        occasions.delete(tournament)


def test_no_orphan_occasions(session, owner):

    activities = ActivityCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True
    )

    tournament = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=activities.add("Sport", username=owner.username),
        period=period
    )

    tournament.bookings.append(Booking(
        username=owner.username,
        last_name='Muster',
        first_name='Peter',
        period_id=period.id
    ))

    session.flush()

    with pytest.raises(sqlalchemy.exc.IntegrityError):
        activities.delete(activities.query().first())


def test_occasion_durations(session, owner):

    activities = ActivityCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)

    retreat = activities.add("Management Retreat", username=owner.username)

    assert not DAYS.has(retreat.durations, DAYS.half)
    assert not DAYS.has(retreat.durations, DAYS.full)
    assert not DAYS.has(retreat.durations, DAYS.many)

    periods.add(
        title="2016",
        prebooking=(datetime(2015, 1, 1), datetime(2015, 12, 31)),
        execution=(datetime(2016, 1, 1), datetime(2016, 12, 31)),
        active=True
    )

    # add an occasion that last half a day
    monday = occasions.add(
        start=datetime(2016, 10, 3, 12),
        end=datetime(2016, 10, 3, 16),
        timezone="Europe/Zurich",
        activity=retreat,
        period=periods.active()
    )
    transaction.commit()

    retreat = activities.query().first()
    assert DAYS.has(retreat.durations, DAYS.half)
    assert not DAYS.has(retreat.durations, DAYS.full)
    assert not DAYS.has(retreat.durations, DAYS.many)

    # add another half a day occasion
    tuesday = occasions.add(
        start=datetime(2016, 10, 4, 12),
        end=datetime(2016, 10, 4, 16),
        timezone="Europe/Zurich",
        activity=retreat,
        period=periods.active()
    )
    transaction.commit()

    retreat = activities.query().first()
    assert DAYS.has(retreat.durations, DAYS.half)
    assert not DAYS.has(retreat.durations, DAYS.full)
    assert not DAYS.has(retreat.durations, DAYS.many)

    # add a full day occasion
    wednesday = occasions.add(
        start=datetime(2016, 10, 5, 8),
        end=datetime(2016, 10, 5, 16),
        timezone="Europe/Zurich",
        activity=retreat,
        period=periods.active()
    )
    transaction.commit()

    retreat = activities.query().first()
    assert DAYS.has(retreat.durations, DAYS.half)
    assert DAYS.has(retreat.durations, DAYS.full)
    assert not DAYS.has(retreat.durations, DAYS.many)

    # add a multi day occasion
    weekend = occasions.add(
        start=datetime(2016, 10, 8, 8),
        end=datetime(2016, 10, 9, 16),
        timezone="Europe/Zurich",
        activity=retreat,
        period=periods.active()
    )
    transaction.commit()

    retreat = activities.query().first()
    assert DAYS.has(retreat.durations, DAYS.half)
    assert DAYS.has(retreat.durations, DAYS.full)
    assert DAYS.has(retreat.durations, DAYS.many)

    # remove a the first half-day occasion (nothing should change)
    occasions.delete(monday)
    transaction.commit()

    retreat = activities.query().first()
    assert DAYS.has(retreat.durations, DAYS.half)
    assert DAYS.has(retreat.durations, DAYS.full)
    assert DAYS.has(retreat.durations, DAYS.many)

    # remove the second half-day occasion (no more half-days)
    occasions.delete(tuesday)
    transaction.commit()

    retreat = activities.query().first()
    assert not DAYS.has(retreat.durations, DAYS.half)
    assert DAYS.has(retreat.durations, DAYS.full)
    assert DAYS.has(retreat.durations, DAYS.many)

    # remove the full day occasion
    occasions.delete(wednesday)
    transaction.commit()

    retreat = activities.query().first()
    assert not DAYS.has(retreat.durations, DAYS.half)
    assert not DAYS.has(retreat.durations, DAYS.full)
    assert DAYS.has(retreat.durations, DAYS.many)

    # remove the remaining occasion
    occasions.delete(weekend)
    transaction.commit()

    retreat = activities.query().first()
    assert not DAYS.has(retreat.durations, DAYS.half)
    assert not DAYS.has(retreat.durations, DAYS.full)
    assert not DAYS.has(retreat.durations, DAYS.many)


def test_occasion_durations_query(session, owner):

    activities = ActivityCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)

    retreat = activities.add("Management Retreat", username=owner.username)
    meeting = activities.add("Management Meeting", username=owner.username)

    periods.add(
        title="2016",
        prebooking=(datetime(2015, 1, 1), datetime(2015, 12, 31)),
        execution=(datetime(2016, 1, 1), datetime(2016, 12, 31)),
        active=True
    )

    # the retreat lasts a weekend
    occasions.add(
        start=datetime(2016, 10, 8, 8),
        end=datetime(2016, 10, 9, 16),
        timezone="Europe/Zurich",
        activity=retreat,
        period=periods.active()
    )

    # the meeting has a day and a half-day occasion
    occasions.add(
        start=datetime(2016, 10, 10, 8),
        end=datetime(2016, 10, 10, 12),
        timezone="Europe/Zurich",
        activity=meeting,
        period=periods.active()
    )

    occasions.add(
        start=datetime(2016, 10, 11, 8),
        end=datetime(2016, 10, 11, 16),
        timezone="Europe/Zurich",
        activity=meeting,
        period=periods.active()
    )

    transaction.commit()

    assert activities.query().count() == 2
    assert activities.for_filter(duration=DAYS.half).query().count() == 1
    assert activities.for_filter(duration=DAYS.full).query().count() == 1
    assert activities.for_filter(duration=DAYS.many).query().count() == 1

    assert activities\
        .for_filter(duration=DAYS.half)\
        .for_filter(duration=DAYS.full)\
        .query().count() == 1

    assert activities\
        .for_filter(duration=DAYS.half)\
        .for_filter(duration=DAYS.full)\
        .for_filter(duration=DAYS.many)\
        .query().count() == 2

    assert activities\
        .for_filter(duration=DAYS.half)\
        .for_filter(duration=DAYS.many)\
        .query().count() == 2


def test_occasion_ages(session, owner):

    activities = ActivityCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)

    retreat = activities.add("Management Retreat", username=owner.username)
    meeting = activities.add("Management Meeting", username=owner.username)

    periods.add(
        title="2016",
        prebooking=(datetime(2015, 1, 1), datetime(2015, 12, 31)),
        execution=(datetime(2016, 1, 1), datetime(2016, 12, 31)),
        active=True
    )

    occasions.add(
        start=datetime(2017, 2, 18, 8),
        end=datetime(2017, 2, 18, 17),
        timezone="Europe/Zurich",
        age=(1, 10),
        activity=retreat,
        period=periods.active()
    )
    occasions.add(
        start=datetime(2018, 2, 19, 8),
        end=datetime(2018, 2, 19, 17),
        timezone="Europe/Zurich",
        age=(10, 20),
        activity=meeting,
        period=periods.active()
    )

    transaction.commit()

    assert activities.for_filter(age_range=(1, 20)).query().count() == 2
    assert activities.for_filter(age_range=(1, 10)).query().count() == 2
    assert activities.for_filter(age_range=(10, 20)).query().count() == 2
    assert activities.for_filter(age_range=(1, 5)).query().count() == 1
    assert activities.for_filter(age_range=(15, 20)).query().count() == 1
    assert activities.for_filter(age_range=(20, 21)).query().count() == 1
    assert activities.for_filter(age_range=(21, 22)).query().count() == 0
    assert activities.for_filter(age_range=(4, 4)).query().count() == 1
    assert activities.for_filter(age_range=(15, 15)).query().count() == 1
    assert activities.for_filter(age_range=(0, 0)).query().count() == 0
    assert activities.for_filter(age_range=(1, 5))\
        .for_filter(age_range=(5, 15)).query().count() == 2

    assert activities.for_filter(age_range=(1, 1)).contains_age_range((0, 2))
    assert activities.for_filter(age_range=(1, 1)).contains_age_range((1, 1))
    assert activities.for_filter(age_range=(1, 1)).contains_age_range((1, 2))

    assert not activities.for_filter(age_range=(1, 1))\
        .contains_age_range((0, 0))

    assert not activities.for_filter(age_range=(1, 10))\
        .contains_age_range((20, 30))

    assert activities\
        .for_filter(age_range=(1, 10))\
        .for_filter(age_range=(20, 30))\
        .contains_age_range((10, 15))

    assert activities\
        .for_filter(age_range=(1, 10))\
        .for_filter(age_range=(20, 30))\
        .contains_age_range((10, 20))

    assert not activities\
        .for_filter(age_range=(1, 10))\
        .for_filter(age_range=(20, 30))\
        .contains_age_range((15, 16))


def test_occasion_owners(session, owner, secondary_owner):

    activities = ActivityCollection(session)

    activities.add("Management Retreat", username=owner.username)
    activities.add("Management Meeting", username=secondary_owner.username)

    transaction.commit()

    assert activities\
        .for_filter(owner='owner@example.org')\
        .query().count() == 1
    assert activities\
        .for_filter(owner='owner@example.org')\
        .for_filter(owner='secondary@example.org')\
        .query().count() == 2

    assert activities\
        .for_filter(owner='secondary@example.org')\
        .query().count() == 1



from enum import IntEnum
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UTCDateTime, UUID
from psycopg2.extras import NumericRange
from sedate import to_timezone
from sqlalchemy import Boolean
from sqlalchemy import case
from sqlalchemy import CheckConstraint
from sqlalchemy import column
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import func
from sqlalchemy import Text
from sqlalchemy.dialects.postgresql import INT4RANGE
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship
from sqlalchemy_utils import aggregated
from uuid import uuid4


class DAYS(IntEnum):
    half = 2**1
    full = 2**2
    many = 2**3

    @staticmethod
    def has(value, mask):
        return value & mask > 0 if value else False


class Occasion(Base, TimestampMixin):
    """ Describes a single occurrence of an Activity. "Occurence" would have
    been a good word for it too, but that's used by onegov.event.

    So occasion it is.

    """

    __tablename__ = 'occasions'

    #: the public id of this occasion
    id = Column(UUID, primary_key=True, default=uuid4)

    #: Timezone of the occasion
    timezone = Column(Text, nullable=False)

    #: Start date and time of the occasion
    start = Column(UTCDateTime, nullable=False)

    #: End date and time of the event (of the first event if recurring)
    end = Column(UTCDateTime, nullable=False)

    #: Describes the location of the activity
    location = Column(Text, nullable=True)

    #: The expected age of participants
    age = Column(
        INT4RANGE, nullable=False, default=NumericRange(6, 17, bounds='[]'))

    #: The expected number of participants
    spots = Column(
        INT4RANGE, nullable=False, default=NumericRange(0, 10, bounds='[]'))

    #: A note about the occurrence
    note = Column(Text, nullable=True)

    #: The activity this occasion belongs to
    activity_id = Column(
        UUID, ForeignKey("activities.id", use_alter=True), nullable=False)

    #: The period this occasion belongs to
    period_id = Column(
        UUID, ForeignKey("periods.id", use_alter=True), nullable=False)

    @aggregated('period', Column(Boolean, default=False))
    def active(self):
        return column('active')

    #: The bookings linked to this occasion
    bookings = relationship(
        'Booking',
        order_by='Booking.created',
        backref='occasion'
    )

    __table_args__ = (
        CheckConstraint('"start" <= "end"', name='start_before_end'),
    )

    @property
    def localized_start(self):
        """ The localized version of the start date/time. """

        return to_timezone(self.start, self.timezone)

    @property
    def localized_end(self):
        """ The localized version of the end date/time. """

        return to_timezone(self.end, self.timezone)

    @hybrid_property
    def duration(self):
        hours = self.duration_in_seconds / 3600

        # defined here and in the expression below!
        if hours <= 6:
            return DAYS.half
        elif hours <= 24:
            return DAYS.full
        else:
            return DAYS.many

    @duration.expression
    def duration(self):

        # defined here and in the property above!
        return case((
            (Occasion.duration_in_seconds <= (6 * 3600), int(DAYS.half)),
            (Occasion.duration_in_seconds <= (24 * 3600), int(DAYS.full)),
        ), else_=int(DAYS.many))

    @hybrid_property
    def duration_in_seconds(self):
        return (self.end - self.start).total_seconds()

    @duration_in_seconds.expression
    def duration_in_seconds(self):
        return func.extract('epoch', self.end - self.start)

    def cancel(self):
        assert self.state in ('active', 'cancelled')
        self.state == 'cancelled'

    def archive(self):
        assert self.state in ('active', 'cancelled', 'archived')
        self.state == 'archived'
